/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		created and modified 2005 ... 2012 by Harald Foerster	   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifndef FILE_H
#define FILE_H

#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_OPENDIR
#include <dirent.h>
#endif

#include "nscache.h"


#ifndef ENODATA
#define ENODATA ENOENT
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define FILE_OPEN_READ_ONLY	(O_RDONLY | O_BINARY)
#define FILE_OPEN_WRITE_ONLY	(O_WRONLY | O_BINARY | O_CREAT | O_EXCL)
#define FILE_OPEN_READ_WRITE	(O_RDWR   | O_BINARY)

#define FILE_ERR_OPEN_SRC	-1
#define FILE_ERR_OPEN_DST	-2
#define FILE_ERR_WRITE		-3
#define FILE_ERR_LINK		-4
#define FILE_ERR_DIR_CREATE	-5

#define FILE_ACCESS_READ	(1 << 0)
#define FILE_ACCESS_WRITE	(1 << 1)
#define FILE_ACCESS_REMOVE	(1 << 2)

#define FILE_PERM_USER 0644
#define FILE_PERM_TEMP 0600
#define DIR_PERMISSION 0755

#define URL_CONVERT_NONE			0
#define URL_CONVERT_ALL_DOMAIN_REFERENCE	1
#define URL_CONVERT_ONLY_FILE_REFERENCE		2


typedef struct
{
	int			folders;
	int			files;
	int			converted;
	int			errors;
	int			errnr;
	char*			fname;
	const nscache_record_t*	erec;

} SaveDirResult;


extern const char	index_html_str[12];
extern unsigned int	decode;
extern unsigned int	single_folder;
extern unsigned int	url_to_local_selector;

#ifdef HAVE_LINK
extern unsigned int	hardlink;
#endif

#ifdef HAVE_OPENDIR
#define DIRdata		DIR
#define DIRopen(path)	opendir(path)
#define DIRread(dir, d)	((d) = readdir(dir), (d) ? (d)->d_name : NULL)
#define DIRclose(dir)	closedir(dir)
#else
#ifdef HAVE_GDIROPEN
#define DIRdata		GDir
#define DIRopen(path)	g_dir_open((path), 0, NULL)
#define DIRread(dir, d)	g_dir_read_name(dir)
#define DIRclose(dir)	g_dir_close(dir)
#endif
#endif

#ifdef BUILD_MINGW
void file_path_dos_to_unix(char *path);
void file_path_unix_to_dos(char *path);
#endif

int  file_access_index(void);
int  file_access(const nscache_record_t* rec);
int  file_chmod(const nscache_record_t* rec, mode_t mode);
int  file_status(const nscache_record_t* rec, struct stat* statbuf);

void *file_read(const nscache_record_t* rec, size_t* length);
void *file_blk_read(const nscache_record_t* rec, size_t* length);
int  file_copy(const nscache_record_t* rec, char* dst, mode_t file_perm);
int  file_save(const nscache_record_t* rec, char* dst);
int  folder_save(const char* dst, const nscache_record_t** rec_array,
							int depth, SaveDirResult* sdr);

/* filename without path */
int  file_open(const char* filename, int flags);

#ifndef CACHE_READONLY
int  file_remove(const char* filename);
int  file_rmdir(const char* dirname);
#endif

#endif /* FILE_H */
