/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005 ... 2012 by Harald Foerster		   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifdef HAVE_MSIE

#include <errno.h>
#include <string.h>
#include <windows.h>
#include <wininet.h>

#include "msie.h"
#include "nls.h"
#include "file.h"

/*
#include "nscache.h"
#include <glib.h>
#include <sys/types.h>
#include <time.h>
#include <winbase.h>
*/

/* not defined in MinGW */
#ifndef ELIBBAD
#define ELIBBAD	EPERM
#endif
#ifndef ELIBACC
#define ELIBACC	EPERM
#endif
#ifndef MAX_CACHE_ENTRY_INFO_SIZE
#define MAX_CACHE_ENTRY_INFO_SIZE 8192
#endif


typedef struct
{
	HMODULE wininet_hModule;
	HANDLE	(WINAPI *FindFirstUrlCacheEntry)
				(LPCSTR, LPINTERNET_CACHE_ENTRY_INFOA, PDWORD);
	BOOL	(WINAPI *FindNextUrlCacheEntry)
				(HANDLE, LPINTERNET_CACHE_ENTRY_INFOA, PDWORD);
	BOOL	(WINAPI *FindCloseUrlCache) (HANDLE);
	HANDLE	(WINAPI *RetrieveUrlCacheEntryStream)
				(LPCSTR, LPINTERNET_CACHE_ENTRY_INFOA, PDWORD, BOOL, DWORD);
	BOOL	(WINAPI *ReadUrlCacheEntryStream)
				(HANDLE, DWORD, LPVOID, PDWORD, DWORD);
	BOOL	(WINAPI *UnlockUrlCacheEntryStream) (HANDLE, DWORD);
	BOOL	(WINAPI *GetUrlCacheEntryInfo)
				(LPCSTR, LPINTERNET_CACHE_ENTRY_INFOA, PDWORD);
#ifndef CACHE_READONLY
	BOOL	(WINAPI *DeleteUrlCacheEntry) (LPCSTR);
#endif

} MSIE_WININET;

static MSIE_WININET msie_wininet = { NULL };
static const char msie_wininet_dll_str[] = "wininet.dll";

static void	*msie_read_db_init(char **dbname, NScacheReadNextFunc *func);
static GSList	*msie_read_db_free(void *ptr);
static int	msie_cache_check(const char *dbname);

#ifndef CACHE_READONLY
static int	msie_record_delete(nscache_record_t *);
#endif


typedef struct
{
	GSList				*slist;
	HANDLE				handle;
	MSIE_WININET			*dll;
	DWORD				error;
	int				max_size;
	LPINTERNET_CACHE_ENTRY_INFOA	info;

} msie_read_data_t;


NSC_BROWSER msie_browser =
{
	msie_read_db_init,
	msie_read_db_free,

#ifndef CACHE_READONLY
	msie_record_delete,
#endif
	msie_cache_check,

	gettext_nop("MSIE")
};



static void msie_set_errno(DWORD error)
{
	int err;

	switch (error)
	{
		case NO_ERROR:
			err = 0;
			break;

		/* No such file or directory */
		case ERROR_FILE_NOT_FOUND:
		case ERROR_PATH_NOT_FOUND:
		case ERROR_DLL_NOT_FOUND:
			err = ENOENT;
			break;

		/* Permission denied */
		case ERROR_ACCESS_DENIED:
			err = EACCES;
			break;

		/* Invalid argument */
		case ERROR_INVALID_PARAMETER:
			err = EINVAL;
			break;

		/* Accessing a corrupted shared lib */
		case ERROR_INVALID_DLL:
			err = ELIBBAD;
			break;

		/* Function not implemented */
		case ERROR_PROC_NOT_FOUND:
			err = ENOSYS;
			break;

		/* Can't access a needed shared lib */
		case ERROR_DLL_INIT_FAILED:
			err = ELIBACC;
			break;

		default:
			err = EPERM;
			break;
	}

	errno = err;

} /* static void msie_set_errno(DWORD error) */

static void msie_close_db(MSIE_WININET *dll)
{
	if (dll->wininet_hModule)
	{
		FreeLibrary(dll->wininet_hModule);

		dll->wininet_hModule = NULL;
	}

} /* static void msie_close_db(MSIE_WININET *dll) */

static MSIE_WININET *msie_open_db(void)
{
	MSIE_WININET *dll = &msie_wininet;

	if (dll->wininet_hModule)
	{
		return dll;
	}

	dll->wininet_hModule = LoadLibraryA(msie_wininet_dll_str);

	if (dll->wininet_hModule)
	{
		do
		{
			if ( (dll->FindFirstUrlCacheEntry =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"FindFirstUrlCacheEntryA")) == NULL )
			{
				break;
			}

			if ( (dll->FindNextUrlCacheEntry =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"FindNextUrlCacheEntryA")) == NULL )
			{
				break;
			}

			if ( (dll->FindCloseUrlCache =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"FindCloseUrlCache")) == NULL )
			{
				break;
			}

			if ( (dll->RetrieveUrlCacheEntryStream =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"RetrieveUrlCacheEntryStreamA")) == NULL )
			{
				break;
			}

			if ( (dll->ReadUrlCacheEntryStream =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"ReadUrlCacheEntryStream")) == NULL )
			{
				break;
			}

			if ( (dll->UnlockUrlCacheEntryStream =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"UnlockUrlCacheEntryStream")) == NULL )
			{
				break;
			}

			if ( (dll->GetUrlCacheEntryInfo =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"GetUrlCacheEntryInfoA")) == NULL )
			{
				break;
			}

#ifndef CACHE_READONLY
			if ( (dll->DeleteUrlCacheEntry =
				(void *) GetProcAddress(dll->wininet_hModule, 
						"DeleteUrlCacheEntry")) == NULL )
			{
				break;
			}
#endif /* ! CACHE_READONLY */

			return dll;
		}
		while (FALSE);

	} /* if (dll->wininet_hModule) */

	msie_set_errno(GetLastError());
	msie_close_db(dll);

	return NULL;

} /* static MSIE_WININET *msie_open_db(void) */

static time_t filetime_to_time_t(const FILETIME *ftm)
{
	if (ftm->dwLowDateTime | ftm->dwHighDateTime)
	{
		FILETIME	ltm;
		SYSTEMTIME	stm;

		if (FileTimeToLocalFileTime(ftm, &ltm))
		{
			ftm = &ltm;
		}

		if (FileTimeToSystemTime(ftm, &stm))
		{
			struct tm	utm;

			utm.tm_isdst = -1;
			utm.tm_sec   = stm.wSecond;
			utm.tm_min   = stm.wMinute;
			utm.tm_hour  = stm.wHour;
			utm.tm_mday  = stm.wDay;

			/* utm: 0 .. 11 -- stm: 1 .. 12 */
			utm.tm_mon   = stm.wMonth - 1;

			if (utm.tm_mon < 0)
			{
				utm.tm_mon = 0;
			}

			/* utm: years after 1900 -- stm: 1601 .. 30827 */
			utm.tm_year = stm.wYear - 1900;

			if (utm.tm_year < 0)
			{
				utm.tm_year = 0;
			}

			return mktime(&utm);
		}
	}

	return 0;

} /* static time_t filetime_to_time_t(const FILETIME *ftm) */

static msie_record_t *msie_dbt_to_record(LPINTERNET_CACHE_ENTRY_INFOA info)
{
	char		*str;
	msie_record_t	*msie_rec;

	if (info->lpszLocalFileName == NULL ||
		info->lpszLocalFileName[0] == '\0' ||
			info->lpszSourceUrlName == NULL ||
				strstr(info->lpszSourceUrlName, "://") == NULL)
	{
		return NULL;
	}

	/* initialize to '\0' */
	msie_rec = g_malloc0(sizeof(msie_record_t));

	msie_rec->record.urlstr = g_strdup(info->lpszSourceUrlName);

	str = strrchr(info->lpszLocalFileName, '\\');

	if (str == NULL)
	{
		msie_rec->record.filename = g_strdup(info->lpszLocalFileName);
		msie_rec->pathname = content_misc_into_gslist(na_str); /* "n/a" */
	}

	else
	{
		str++;
		msie_rec->record.filename = g_strdup(str);

		/* always be nul-terminated */
		info->lpszLocalFileName[str - info->lpszLocalFileName] = '\0';
		msie_rec->pathname = content_misc_into_gslist(info->lpszLocalFileName);
	}

	msie_rec->record.last_modified = filetime_to_time_t(&info->LastModifiedTime);
	msie_rec->record.last_accessed = filetime_to_time_t(&info->LastAccessTime);
	msie_rec->record.expires = filetime_to_time_t(&info->ExpireTime);

	msie_rec->record.content_length = info->dwSizeLow;

	/* avoid accessing files directly - use always 'wininet.dll' */
	msie_rec->record.content_offset = MSIE_CACHE;

	/* Content-Type: */
	str = tl_get_mime_value_str(info->lpHeaderInfo, info->dwHeaderInfoSize,
					"content-type:", sizeof("content-type:") - 1);
	msie_rec->record.content_type = content_type_into_glist(str);

	/* charset= */
	str = tl_get_mime_value_str(info->lpHeaderInfo, info->dwHeaderInfoSize,
						"charset=", sizeof("charset=") - 1);
	msie_rec->record.charset = content_misc_into_gslist(str);

#if 0
	/* seems no entry for "Content-Encoding:" */
	str = tl_get_mime_value_str(info->lpHeaderInfo, info->dwHeaderInfoSize,
				"content-encoding:", sizeof("content-encoding:") - 1);
	msie_rec->record.content_encoding = content_misc_into_gslist(str);

#endif /* 0 */

	return msie_rec;

} /* static msie_record_t *msie_dbt_to_record(info) */

static int msie_read_db_next(void *ptr)
{
	msie_record_t		*msie_rec;
	msie_read_data_t	*data;

	data = ptr;

	for (;;)
	{
		DWORD error, info_size;

		info_size = data->max_size;

		if (data->handle == NULL)
		{
			/* "cookie:", "visited:"  or NULL ("*.*) */
			data->handle = (data->dll)->FindFirstUrlCacheEntry(NULL,
								data->info, &info_size);
			if (data->handle)
			{
				break;
			}

		}

		else if ((data->dll)->FindNextUrlCacheEntry(data->handle,
								data->info, &info_size))
		{
				break;
		}

		error = GetLastError();

		if (error != ERROR_INSUFFICIENT_BUFFER)
		{
			data->error = error;
			return FALSE;
		}

		data->max_size	= info_size;
		data->info	= g_realloc(data->info, data->max_size);

	} /* for (;;) */

	msie_rec = msie_dbt_to_record(data->info);

	if (msie_rec)
	{
		data->slist = g_slist_prepend(data->slist, msie_rec);
	}

	return TRUE;

} /* static int msie_read_db_next(void*) */

static void *msie_read_db_init(char **dbname, NScacheReadNextFunc *func)
{
	MSIE_WININET		*dll;
	msie_read_data_t	*data;

	*func	= (NScacheReadNextFunc) msie_read_db_next;
	*dbname	= g_strdup(msie_wininet_dll_str);

	dll = msie_open_db();

	if (dll == NULL)
	{
		gui_strerror(gettext_nop("Error opening 'wininet.dll'"));
		return NULL;
	}

	data = g_malloc(sizeof(msie_read_data_t));

	data->slist	= NULL;
	data->handle	= NULL;
	data->dll	= dll;
	data->error	= 0;
	data->max_size	= MAX_CACHE_ENTRY_INFO_SIZE;
	data->info	= g_malloc(MAX_CACHE_ENTRY_INFO_SIZE);

	return data;

} /* static void *msie_read_db_init(char**, NScacheReadNextFunc*) */

static GSList *msie_read_db_free(void *ptr)
{
	GSList			*retv;
	msie_read_data_t	*data;

	data = ptr;

	if (data->handle)
	{
		(data->dll)->FindCloseUrlCache(data->handle);
	}

	if (data->error && data->error != ERROR_NO_MORE_ITEMS)
	{
		msie_set_errno(data->error);
		gui_strerror(gettext_nop("Error reading 'wininet.dll'"));
	}

	else
	{
		gui_strerror(NULL);
	}

	retv = data->slist;

	g_free(data->info);
	g_free(data);

	return retv;

} /* static GSList *msie_read_db_free(void*) */

#ifndef CACHE_READONLY

static int msie_record_delete(nscache_record_t *nsc_rec)
{
	if (msie_wininet.wininet_hModule &&
		msie_wininet.DeleteUrlCacheEntry(nsc_rec->urlstr) == FALSE)
	{
		msie_set_errno(GetLastError());

		return NSC_ERR_ENTRY;
	}

	return 0;

} /* static int msie_record_delete(nscache_record_t *) */

#endif /* ! CACHE_READONLY */

static int msie_cache_check(const char *dummy)
{
	MSIE_WININET *dll = msie_open_db();

	if (dll)
	{
		msie_close_db(dll);

		return 0;
	}

	return -1;
}

int msie_file_access(const nscache_record_t *nsc_rec)
{
	DWORD				info_size;
	LPINTERNET_CACHE_ENTRY_INFOA	info;
	char				*mem;
	char				infobuf[MAX_CACHE_ENTRY_INFO_SIZE];

	mem	  = NULL;
	info_size = sizeof(infobuf);
	info	  = (LPINTERNET_CACHE_ENTRY_INFOA) &infobuf;


	while (msie_wininet.GetUrlCacheEntryInfo(nsc_rec->urlstr,
						info, &info_size) == FALSE)
	{
		DWORD error = GetLastError();

		if (error == ERROR_INSUFFICIENT_BUFFER)
		{
			void *ptr = realloc(mem, info_size);

			if (ptr)
			{
				mem  = ptr;
				info = ptr;

				continue;
			}
		}

		if (mem)
		{
			free(mem);
		}

		return -1;
	}

	if (mem)
	{
		free(mem);
	}

#ifndef CACHE_READONLY
	return FILE_ACCESS_READ | FILE_ACCESS_WRITE | FILE_ACCESS_REMOVE;
#else
	return FILE_ACCESS_READ;
#endif

} /* int msie_file_access(const nscache_record_t *nsc_rec) */

void *msie_file_read(const nscache_record_t *nsc_rec, size_t *length)
{
	HANDLE				handle;
	DWORD				error;
	LPINTERNET_CACHE_ENTRY_INFOA	info;
	char				*mem;
	char				infobuf[MAX_CACHE_ENTRY_INFO_SIZE];

	union
	{
		DWORD	info_size;
		DWORD	file_size;
	} u;

	mem	  	= NULL;
	u.info_size	= sizeof(infobuf);
	info	  	= (LPINTERNET_CACHE_ENTRY_INFOA) &infobuf;

	while ( (handle = msie_wininet.RetrieveUrlCacheEntryStream(nsc_rec->urlstr,
						info, &u.info_size, FALSE, 0)) == NULL )
	{
		error = GetLastError();

		if (error == ERROR_INSUFFICIENT_BUFFER)
		{
			void *ptr = realloc(mem, u.info_size);

			if (ptr)
			{
				mem  = ptr;
				info = ptr;

				continue;
			}
		}

		if (mem)
		{
			free(mem);
		}

		msie_set_errno(error);

		return NULL;
	}


	u.file_size = info->dwSizeLow;

	if (mem)
	{
		free(mem);
	}

	mem = malloc(u.file_size + 1);

	if (mem == NULL)
	{
		return mem;
	}


	while (error = 0, msie_wininet.ReadUrlCacheEntryStream(handle,
					0, mem, &u.file_size, 0) == FALSE)
	{
		error = GetLastError();

		if (error == ERROR_INSUFFICIENT_BUFFER)
		{
			void *ptr = realloc(mem, u.file_size + 1);

			if (ptr)
			{
				mem = ptr;

				continue;
			}
		}

		break;
	}


	msie_wininet.UnlockUrlCacheEntryStream(handle, 0);

	if (error)
	{
		free(mem);

		msie_set_errno(error);

		return NULL;
	}

	*length = u.file_size;
	mem[*length] = '\0';

	return mem;

} /* void *msie_file_read(const nscache_record_t *, size_t *) */

#endif /* HAVE_WININET_H */

/* EOF */
