/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		created and modified 2004,2005,2009 by Harald Foerster	   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifndef STRINGBUF_H
#define STRINGBUF_H

#include <stdlib.h>


enum
{
	STRINGBUF_INIT_SIZE	= 512
};


/* Not In List */
#define NIL	((void*) -1)

typedef struct
{
	char*		str;	/* string buffer */
	size_t		len;	/* string length */
	size_t		sz;	/* including '\0' (len+1) */
	size_t		max;	/* max buffer size */
	void*		ptr;	/* if memory for this structure was malloc'd */

} StringBuf;


/* allocate memory for the string and the structure (if 'buffer' is NULL) */
StringBuf* stringbuf_new(StringBuf* buffer, size_t size);

/* call 'stringbuf_new' and 'stringbuf_copy' */
StringBuf* stringbuf_init(StringBuf* buffer, const char* string, int length);

/* make a stringbuf from string ('string' must be malloc'ed memory) */
StringBuf* stringbuf_from_string(StringBuf* buffer, char* string, int length);

/* free allocated memory and returns always NULL */
StringBuf* stringbuf_delete(StringBuf* buffer);

/* enlarge string memory */
char* stringbuf_realloc(StringBuf* buffer, size_t size);

/* move 'buffer->str' from 'position' to 'position + length' (length < 0 moves down) */
char* stringbuf_move(StringBuf* buffer, size_t position, int length);

/* copy 'string' at start of string memory */
char* stringbuf_copy(StringBuf* buffer, const char* string, int length);

/* append 'string' at end */
char* stringbuf_append(StringBuf* buffer, const char* string, int length);

/* insert 'string' at 'position' */
char* stringbuf_insert(StringBuf* buffer, const char* string, int length, size_t position);

/* replace first occurence of 's1' with 's2' and return start of 's1' (NIL not found, NULL no memory) */
char* stringbuf_replace(StringBuf* buffer, const char* s1, int l1, const char* s2, int l2);

/* replace last occurence of 's1' with 's2' and return start of 's1' (NIL not found, NULL no memory) */
char* stringbuf_replace_last(StringBuf* buffer, const char* s1, int l1, const char* s2, int l2);

/* shorten the string to 'position' */
char* stringbuf_truncate(StringBuf* buffer, size_t position);

/* append the character at end */
char* stringbuf_character_append(StringBuf* buffer, int ch);

/* replace character with 'ch' at string 'position' */
char* stringbuf_character_replace(StringBuf* buffer, int ch, size_t position);

/* insert character 'ch' at string 'position' */
char* stringbuf_character_insert(StringBuf* buffer, int ch, size_t position);


/* insert 'string' at beginning of buffer */
#define stringbuf_prepend(buffer, string, length)	\
		stringbuf_insert((buffer), (string), (length), 0)

/* remove first occurence of 'string' and return start of 's1' (or NULL if none found ) */
#define stringbuf_remove(buffer, string, length)	\
		stringbuf_replace((buffer), (string), (length), NULL, 0)

/* remove first occurence of 'string' and return start of 's1' (or NULL if none found ) */
#define stringbuf_remove_last(buffer, string, length)	\
		stringbuf_replace_last((buffer), (string), (length), NULL, 0)


#endif /* STRINGBUF_H */
